--style=whitesmith
--indent=spaces=3
--attach-namespaces
--break-one-line-headers
--convert-tabs
--max-code-length=120
--indent-preproc-block
#--pad-oper
--pad-comma
--unpad-paren
--align-pointer=type
--align-reference=type
-j # --add-braces / --add-brackets
-J # --add-one-line-braces / --add-one-line-brackets
